import re
import time
from dataclasses import dataclass
from typing import List

from oaklib.interfaces.semsim_interface import SemanticSimilarityInterface
from oaklib.selector import get_adapter
from linkml_runtime.dumpers.json_dumper import JSONDumper
from loguru import logger
import pystow

from monarch_py.datamodels.model import TermSetPairwiseSimilarity
from monarch_py.implementations.oak.annotation_utils import (
    get_word_length,
    concatenate_same_entities,
    concatenate_ngram_entities,
    replace_entities,
    convert_to_json,
)


@dataclass
class OakImplementation(SemanticSimilarityInterface):
    """Implementation of Monarch Interfaces for OAK"""

    semsim = None
    json_dumper = JSONDumper()
    default_predicates = ["rdfs:subClassOf", "BFO:0000050", "UPHENO:0000001"]

    default_phenio_db_url = "https://data.monarchinitiative.org/monarch-kg-dev/latest/phenio.db.gz"
    phenio_adapter = None

    def init_phenio_adapter(self, phenio_path: str = None, force_update: bool = False):
        if self.phenio_adapter is None:
            logger.info("Warming up semsimian")
            start = time.time()
            # self.phenio_adapter = get_adapter(f"sqlite:obo:phenio")

            if phenio_path:
                logger.debug(f"Creating phenio adapter using phenio_path at {phenio_path}")
                self.phenio_adapter = get_adapter(f"sqlite:{phenio_path}")
            else:
                monarchstow = pystow.module("monarch")

                with monarchstow.ensure_gunzip(
                    "phenio", url=self.default_phenio_db_url, force=force_update
                ) as stowed_phenio_path:
                    logger.debug(f"Creating phenio adapter using pystow at {stowed_phenio_path}")
                    self.phenio_adapter = get_adapter(f"sqlite:{stowed_phenio_path}")

            # run a query to get the adapter to initialize properly
            logger.debug("Running query to initialize adapter")

            # TODO: run a little bit of text annotation here to get oak warmed up
            print("Oaklib adapter warmup")
            for ann in self.phenio_adapter.annotate_text("Nystagmus, strabismus and fundus."):
                print(ann.subject_start, ann.subject_end, ann.object_id, ann.object_label)

            logger.info(f"Phenio adapter ready, warmup time: {time.time() - start} sec")
            return self

    def init_semsim(self, phenio_path: str = None, force_update: bool = False):
        if self.semsim is None:
            logger.info("Warming up semsimian")
            start = time.time()
            # self.semsim = get_adapter(f"sqlite:obo:phenio")

            if phenio_path:
                logger.debug(f"Creating semsimian adapter using phenio_path at {phenio_path}")
                self.semsim = get_adapter(f"semsimian:sqlite:{phenio_path}")
            else:
                monarchstow = pystow.module("monarch")

                with monarchstow.ensure_gunzip(
                    "phenio", url=self.default_phenio_db_url, force=force_update
                ) as stowed_phenio_path:
                    logger.debug(f"Creating semsimian adapter using pystow at {stowed_phenio_path}")
                    self.semsim = get_adapter(f"semsimian:sqlite:{stowed_phenio_path}")

            # run a query to get the adapter to initialize properly
            logger.debug("Running query to initialize adapter")
            self.semsim.termset_pairwise_similarity(
                subjects=["MP:0010771"],
                objects=["HP:0004325"],
                predicates=self.default_predicates,
                labels=False,
            )
            logger.info(f"Semsimian ready, warmup time: {time.time() - start} sec")
            return self

    def compare(
        self, subjects: List[str], objects: List[str], predicates: List[str] = None, labels=False
    ) -> TermSetPairwiseSimilarity:
        """Compare two sets of terms using OAK"""
        predicates = predicates or self.default_predicates
        logger.debug(f"Comparing {subjects} to {objects} using {predicates}")
        compare_time = time.time()
        response = self.semsim.termset_pairwise_similarity(  # type: ignore
            subjects=subjects,
            objects=objects,
            predicates=predicates,
            labels=labels,
        )
        logger.debug(f"Comparison took: {time.time() - compare_time} sec")

        response_dict = self.json_dumper.to_dict(response)
        return TermSetPairwiseSimilarity(**response_dict)

    def annotate_text(self, text):
        """Annotate text using OAK"""
        paragraphs = text.split("\n")
        paragraphs_annotated = []
        for paragraph in paragraphs:
            result = ""
            sentences = re.split(r"(?<!\w\.\w.)(?<![A-Z][a-z]\.)(?<=\.|\?)\s", paragraph)
            for sentence in sentences:
                entities = []
                for ann in self.phenio_adapter.annotate_text(sentence):  # type: ignore
                    if len(ann.object_label) >= 4:
                        element = [ann.subject_start, ann.subject_end, str(ann.object_label) + "," + str(ann.object_id)]
                        if (get_word_length(sentence, ann.subject_start - 1) - len(ann.object_label)) < 2:
                            entities.append(element)
                try:
                    entities.sort()
                    entities = concatenate_same_entities(entities)
                    entities = concatenate_ngram_entities(entities)
                    replaced_text = replace_entities(sentence, entities)
                    result += replaced_text + " "
                except IndexError as error:
                    # Handling the list index out of range error
                    result += sentence + " "
                    print("Error occurred:", error)

            paragraphs_annotated.append(result)
        result = convert_to_json(paragraphs_annotated)
        return result



from abc import ABC
from typing import List
from monarch_py.datamodels.model import TextAnnotationResult


class TextAnnotatorInterface(ABC):
    def get_annotated_entities(self, text: str) -> List[TextAnnotationResult]:
        raise NotImplementedError

    def annotate_text(self, text: str) -> str:
        raise NotImplementedError



import uvicorn
from fastapi import FastAPI, Query
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import RedirectResponse

from monarch_py.api import association, entity, histopheno, search, semsim, text_annotation
from monarch_py.api.config import semsimian, spacyner
from monarch_py.api.middleware.logging_middleware import LoggingMiddleware
from monarch_py.utils.utils import get_release_metadata, get_release_versions

PREFIX = "/v3/api"

app = FastAPI(
    docs_url="/v3/docs",
    redoc_url="/v3/redoc",
)


@app.on_event("startup")
async def initialize_app():
    semsimian()
    spacyner()
    # oak()


app.include_router(association.router, prefix=f"{PREFIX}/association")
app.include_router(entity.router, prefix=f"{PREFIX}/entity")
app.include_router(histopheno.router, prefix=f"{PREFIX}/histopheno")
app.include_router(search.router, prefix=PREFIX)
app.include_router(semsim.router, prefix=f"{PREFIX}/semsim")
app.include_router(text_annotation.router, prefix=PREFIX)

# Allow CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(LoggingMiddleware)

app.description = """

# This is the v3 Monarch API.

This API is a RESTful web service that provides programmatic access to the Monarch Initiative's knowledge graph,
and which serves as the backend to the [Monarch Initiative's website](https://monarchinitiative.org).
"""


@app.get("/")
async def _root():
    return RedirectResponse(url="/v3/docs")


@app.get("/api")
async def _api():
    return RedirectResponse(url="/v3/docs")


@app.get(f"{PREFIX}/releases")
async def _v3(
    dev: bool = Query(default=False, title="Get dev releases of the KG (default False)"),
    limit: int = Query(default=0, title="The number of releases to return (default 0 for no limit)"),
    release: str = Query(default=None, title="Get metadata for a specific release"),
):
    if release is None:
        return get_release_versions(dev=dev, limit=limit)
    return get_release_metadata(release=release, dev=dev)


def run():
    uvicorn.run("monarch_py.api.main:app", host="127.0.0.1", port=8000, reload=True)


if __name__ == "__main__":
    run()



from io import StringIO
from typing import List, Union

from fastapi import APIRouter, Depends, HTTPException, Path, Query, Response
from fastapi.responses import StreamingResponse

from monarch_py.api.additional_models import PaginationParams
from monarch_py.api.config import solr
from monarch_py.api.additional_models import OutputFormat
from monarch_py.datamodels.model import AssociationTableResults, Node
from monarch_py.datamodels.category_enums import AssociationCategory
from monarch_py.utils.format_utils import to_json, to_tsv

router = APIRouter(tags=["entity"], responses={404: {"description": "Not Found"}})


@router.get("/{id}")
async def _get_entity(
    id: str = Path(
        title="ID of the entity to retrieve",
        examples=["MONDO:0019391"],
    ),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
) -> Union[Node, str]:
    """Retrieves the entity with the specified id

    <b>Args:</b> <br>
        id (str): ID for the entity to retrieve, ex: MONDO:0019391

    <b>Raises:</b> <br>
        HTTPException: 404 if the entity is not found

    <b>Returns:</b> <br>
        Node: Entity details for the specified id
    """
    response = solr().get_entity(id, extra=True)
    if response is None:
        raise HTTPException(status_code=404, detail="Entity not found")
    if format == OutputFormat.json:
        return Node(**response.__dict__)  # This is an odd consequence of how Node extends Entity
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")


@router.get("/{id}/{category}")
def _association_table(
    id: str = Path(
        title="ID of the entity to retrieve association table data for",
        examples=["MONDO:0019391"],
    ),
    category: AssociationCategory = Path(
        title="Type of association to retrieve association table data for",
        examples=["biolink:DiseaseToPhenotypicFeatureAssociation"],
    ),
    query: str = Query(default=None, title="query string to limit results to a subset", examples=["thumb"]),
    traverse_orthologs: bool = Query(
        default=False,
        title="Traverse orthologs to get associations",
        examples=[True, False],
    ),
    sort: List[str] = Query(
        default=None,
        title="Sort results by a list of field + direction statements",
        examples=["subject_label asc", "predicate asc", "object_label asc"],
    ),
    pagination: PaginationParams = Depends(),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
    download: bool = Query(
        default=False,
        title="Download the results as a file",
        examples=[True, False],
    ),
    direct: bool = Query(
        default=False,
        title="Only return direct associations",
        examples=[True, False],
    ),
    facet_fields: List[str] = Query(
        default=None,
        title="Facet fields to include in the response",
        examples=["subject", "subject_taxon", "predicate"],
    ),
    facet_queries: List[str] = Query(
        default=None, title="Facet queries to include in the response", examples=['subject_category:"biolink:Gene"']
    ),
    filter_queries: List[str] = Query(
        default=None, title="Filter queries to limit the response", examples=['subject_category:"biolink:Gene"']
    ),
) -> Union[AssociationTableResults, str]:
    """
    Retrieves association table data for a given entity and association type

    <b>Args:</b> <br>
        id (str): ID of the entity to retrieve association table data, ex: MONDO:0019391 <br>
        category (str): Category of association to retrieve association table data for, ex: biolink:DiseaseToPhenotypicFeatureAssociation <br>
        Path (str, optional): Path string to limit results to a subset. Defaults to None. <br>
        pagination (PaginationParams, optional): Pagination parameters. Defaults to Depends(). <br>

    <b>Returns:</b> <br>
        AssociationResults: Association table data for the specified entity and association type
    """
    response = solr().get_association_table(
        entity=id,
        category=category,
        q=query,
        traverse_orthologs=traverse_orthologs,
        direct=direct,
        facet_fields=facet_fields,
        facet_queries=facet_queries,
        filter_queries=filter_queries,
        sort=sort,
        offset=pagination.offset,
        limit=pagination.limit,
    )
    if download is True:
        string_response = (
            to_tsv(response, print_output=False)
            if format == OutputFormat.tsv
            else to_json(response, print_output=False)
        )
        stream = StringIO(string_response)
        response = StreamingResponse(
            stream, media_type="text/tab-separated-values" if format == OutputFormat.tsv else "application/json"
        )
        response.headers["Content-Disposition"] = f"attachment; filename=assoc-table-{id}.{format.value}"
        return response
    if format == OutputFormat.json:
        return response
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")



from typing import List, Union

from fastapi import APIRouter, Depends, Query, Response

from monarch_py.api.additional_models import OutputFormat, PaginationParams
from monarch_py.api.config import solr
from monarch_py.datamodels.model import SearchResults
from monarch_py.datamodels.category_enums import EntityCategory, MappingPredicate
from monarch_py.utils.format_utils import to_tsv

router = APIRouter(
    tags=["search"],
    responses={404: {"description": "Not Found"}},
)


@router.get("/search")
async def search(
    q: str = Query(default=None),
    category: Union[List[EntityCategory], None] = Query(default=None),
    in_taxon_label: Union[List[str], None] = Query(default=None),
    pagination: PaginationParams = Depends(),
) -> SearchResults:
    """Search for entities by label, with optional filters

    Args:
        q (str, optional): Query string. Defaults to "*:*".
        category (str, optional): Filter by biolink model category. Defaults to None.
        in_taxon_label (str, optional): Filter by taxon label. Defaults to None.
        offset (int, optional): Offset for pagination. Defaults to 0.
        limit (int, optional): Limit results. Defaults to 20.

    Returns:
        EntityResults
    """
    facet_fields = ["category", "in_taxon_label"]
    if category is None:
        category = []
    response = solr().search(
        q=q or "*:*",
        category=category,
        in_taxon_label=in_taxon_label,
        facet_fields=facet_fields,
        offset=pagination.offset,
        limit=pagination.limit,
    )

    return response


@router.get("/autocomplete")
async def autocomplete(
    q: str = Query(
        default="*:*",
        title="Query string to autocomplete against",
        examples=["fanc", "ehler"],
    )
) -> SearchResults:
    """Autocomplete for entities by label

    Args:
        q (str): Query string to autocomplete against

    Returns:
        SearchResults
    """
    response = solr().autocomplete(q=q)
    return response


@router.get("/mappings")
async def mappings(
    entity_id: Union[List[str], None] = Query(default=None),
    subject_id: Union[List[str], None] = Query(default=None),
    predicate_id: Union[List[MappingPredicate], None] = Query(default=None),
    object_id: Union[List[str], None] = Query(default=None),
    mapping_justification: Union[List[str], None] = Query(default=None),
    pagination: PaginationParams = Depends(),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
):
    response = solr().get_mappings(
        entity_id=entity_id,
        subject_id=subject_id,
        predicate_id=predicate_id,
        object_id=object_id,
        mapping_justification=mapping_justification,
        offset=pagination.offset,
        limit=pagination.limit,
    )
    if format == OutputFormat.json:
        return response
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")



from abc import ABC, abstractmethod

from monarch_py.datamodels.model import EntityResults


class QueryInterface(ABC):
    """Abstract interface for querying the Monarch KG in a SQL way"""

    @abstractmethod
    def query(self) -> EntityResults:
        """

        Args:
            q (str): Query string to match against
            category (str): Limit results to only this category
            taxon (str): Limit results to only this taxon
            offset (int): Offset of the first result to return, defaults to 0
            limit (int): Limit the number of results to return, defaults to 20

        Raises:
            NotImplementedError: Use a specific implementation (see the documentation for a list of implementations)

        Returns:
            EntityResults: Dataclass representing results of a generic entity search.
        """
        raise NotImplementedError



from typing import List, Optional
from typing_extensions import Annotated

import typer

from monarch_py.datamodels.category_enums import (
    AssociationCategory,
    AssociationPredicate,
)
from monarch_py.implementations.sql.sql_implementation import SQLImplementation
from monarch_py.utils.utils import console, set_log_level
from monarch_py.utils.format_utils import format_output
from monarch_py.utils import cli_fields as fields

sql_app = typer.Typer()
app_state = {"log_level": "WARNING"}


@sql_app.callback(invoke_without_command=True)
def callback(
    ctx: typer.Context,
    quiet: Annotated[bool, typer.Option("--quiet", "-q", help="Set log level to warning")] = False,
    debug: Annotated[bool, typer.Option("--debug", "-d", help="Set log level to debug")] = False,
):
    if ctx.invoked_subcommand is None:
        typer.secho(
            f"\n\tNo command specified\n\tTry `monarch sql --help` for more information.\n",
            fg=typer.colors.YELLOW,
        )
        raise typer.Exit()
    log_level = "DEBUG" if debug else "WARNING" if quiet else "INFO"
    set_log_level(log_level)


@sql_app.command()
def entity(
    entity_id: Annotated[
        str,
        typer.Argument(help="The identifier of the entity to be retrieved"),
    ],
    extra: Annotated[
        bool,
        typer.Option(
            "--extra",
            "-e",
            help="Include extra fields in the output (association_counts and node_hierarchy)",
        ),
    ],
    update: Annotated[
        bool,
        typer.Option(
            "--update",
            "-u",
            help="Whether to re-download the Monarch KG",
        ),
    ],
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """Retrieve an entity by ID"""
    data = SQLImplementation()
    response = data.get_entity(entity_id, update, extra)

    if not response:
        console.print(f"\nEntity '{id}' not found.\n")
        raise typer.Exit(1)
    format_output(fmt, response, output)


@sql_app.command()
def associations(
    category: Annotated[
        Optional[List[AssociationCategory]],
        typer.Option(
            "--category",
            "-c",
            help="Category to get associations for",
        ),
    ] = None,
    subject: Annotated[
        Optional[List[str]],
        typer.Option(
            "--subject",
            "-s",
            help="Subject ID to get associations for",
        ),
    ] = None,
    predicate: Annotated[
        Optional[List[AssociationPredicate]],
        typer.Option(
            "--predicate",
            "-p",
            help="Predicate ID to get associations for",
        ),
    ] = None,
    object: Annotated[
        Optional[List[str]],
        typer.Option(
            "--object",
            "-o",
            help="Object ID to get associations for",
        ),
    ] = None,
    entity: Annotated[
        Optional[List[str]],
        typer.Option(
            "--entity",
            "-e",
            help="Entity (subject or object) ID to get associations for",
        ),
    ] = None,
    direct: Annotated[
        bool,
        typer.Option(
            "--direct",
            "-d",
            help="Whether to exclude associations with subject/object as ancestors",
        ),
    ] = False,
    compact: Annotated[
        bool,
        typer.Option(
            "--compact",
            "-C",
            help="Whether to return a compact representation of the associations",
        ),
    ] = False,
    limit: fields.LimitOption = 20,
    offset: fields.OffsetOption = 0,
    fmt: fields.FormatOption = fields.OutputFormat.json,
    output: fields.OutputOption = None,
):
    """
    Paginate through associations
    """
    args = locals()
    args.pop("fmt", None)
    args.pop("output", None)

    data = SQLImplementation()
    response = data.get_associations(**args)
    format_output(fmt, response, output)



"""
Utility functions for annotating text with OAK.
"""

import re
import json
from typing import List


def get_word_length(text, start):
    word = ""
    index = start
    while index < len(text) and text[index].isalpha():
        word += text[index]
        index += 1
    return len(word)


def concatenate_same_entities(lst):
    result = {}
    for elem in lst:
        key = (elem[0], elem[1])
        if key in result:
            result[key] += "|" + elem[2]
        else:
            result[key] = elem[2]
    concatenated_list = [[key[0], key[1], value] for key, value in result.items()]
    return concatenated_list


def concatenate_ngram_entities(lst):
    merged_list = []
    start, end, text = lst[0]
    for element in lst[1:]:
        if element[0] <= end:  # Check if range overlaps
            end = max(end, element[1])  # Merge the range
            text += "|" + element[2]  # Concatenate the texts
        else:
            merged_list.append([start, end, text])  # Add the merged element to the result
            start, end, text = element  # Move to the next element
    merged_list.append([start, end, text])  # Add the last merged element
    return merged_list


def replace_entities(text, entities):
    replaced_text = text
    # Sort the entities in descending order of start character indices
    entities = sorted(entities, key=lambda x: x[0], reverse=True)
    for entity in entities:
        start, end = entity[0] - 1, entity[1]
        # entity_value = entity[2]
        entity_value = f'<span class="sciCrunchAnnotation" data-sciGraph="{entity[2]}">{text[start:end]}</span>'
        replaced_text = replaced_text[:start] + entity_value + replaced_text[end:]
    return replaced_text


def convert_to_json(paragraphs: List[str]):
    result = []
    span_pattern = re.compile(r'<span class="sciCrunchAnnotation" data-sciGraph="([^"]+)">([^<]+)</span>')

    for paragraph in paragraphs:
        start_index = 0
        for match in span_pattern.finditer(paragraph):
            span_data = match.group(1)
            span_text = match.group(2)

            if start_index < match.start():
                non_span_text = paragraph[start_index : match.start()]
                result.append({"text": non_span_text})

            tokens = []
            for token_data in span_data.split("|"):
                token_parts = token_data.split(",")
                tokens.append({"id": token_parts[1], "name": token_parts[0]})

            result.append({"text": span_text, "tokens": tokens})
            start_index = match.end()

        if start_index < len(paragraph):
            non_span_text = paragraph[start_index:]
            result.append({"text": non_span_text})

        result.append({"text": "\n"})

    api_response = json.dumps(result)
    # Load the JSON
    data = json.loads(api_response)

    return data



import json
from typing import Dict, List

import requests
from loguru import logger
from monarch_py.datamodels.solr import SolrQuery, SolrQueryResult, core
from monarch_py.utils.utils import escape
from pydantic import BaseModel


class SolrService(BaseModel):
    base_url: str
    core: core

    def get(self, id):
        url = f"{self.base_url}/{self.core.value}/get?id={id}"
        response = requests.get(url)
        response.raise_for_status()
        entity = response.json()["doc"]
        try:
            self._strip_json(entity, "_version_", "iri")
        except TypeError:  # if entity is None
            return None
        return entity

    def query(self, q: SolrQuery) -> SolrQueryResult:
        url = f"{self.base_url}/{self.core.value}/select?{q.query_string()}"
        response = requests.get(url)
        logger.debug(f"SolrService.query: {url}")
        data = json.loads(response.text)
        if "error" in data:
            logger.error("Solr error message: " + data["error"]["msg"])
        response.raise_for_status()
        solr_query_result = SolrQueryResult.model_validate(data, from_attributes=True)
        for doc in solr_query_result.response.docs:
            self._strip_json(
                doc,
                "_version_",
                "iri",
                "frequency_computed_sortable_float",
                "has_quotient_sortable_float",
                "has_percentage_sortable_float",
            )

        return solr_query_result

    def _strip_json(self, doc: dict, *fields_to_remove: str):
        for field in fields_to_remove:
            try:
                del doc[field]
            except KeyError:
                pass
        return doc

    # Solr returns facet values and counts as a list, they make much more
    # sense as a dictionary
    def _facets_to_dict(self, facet_list: List[str]) -> Dict:
        return dict(zip(facet_list[::2], facet_list[1::2]))

    def get_filtered_facet(self, id, filter_field, facet_field):
        query = SolrQuery(
            rows=0,
            facet=True,
            facet_fields=[facet_field],
            filter_queries=[f"{filter_field}:{escape(id)}"],
        )

        result = self.query(query)

        facet_fields = result.facet_counts.facet_fields[facet_field]

        return self._facets_to_dict(facet_fields)



from enum import Enum


class EntityCategory(Enum):
    """Entity categories"""

    SEQUENCE_VARIANT = "biolink:SequenceVariant"
    GENE = "biolink:Gene"
    GENOTYPE = "biolink:Genotype"
    PHENOTYPIC_FEATURE = "biolink:PhenotypicFeature"
    BIOLOGICAL_PROCESS_OR_ACTIVITY = "biolink:BiologicalProcessOrActivity"
    DISEASE = "biolink:Disease"
    GROSS_ANATOMICAL_STRUCTURE = "biolink:GrossAnatomicalStructure"
    CELL = "biolink:Cell"
    PATHWAY = "biolink:Pathway"
    NAMED_THING = "biolink:NamedThing"
    ANATOMICAL_ENTITY = "biolink:AnatomicalEntity"
    CELLULAR_COMPONENT = "biolink:CellularComponent"
    MOLECULAR_ENTITY = "biolink:MolecularEntity"
    BIOLOGICAL_PROCESS = "biolink:BiologicalProcess"
    MACROMOLECULAR_COMPLEX = "biolink:MacromolecularComplex"
    MOLECULAR_ACTIVITY = "biolink:MolecularActivity"
    PROTEIN = "biolink:Protein"
    CELLULAR_ORGANISM = "biolink:CellularOrganism"
    VERTEBRATE = "biolink:Vertebrate"
    VIRUS = "biolink:Virus"
    BEHAVIORAL_FEATURE = "biolink:BehavioralFeature"
    CHEMICAL_ENTITY = "biolink:ChemicalEntity"
    LIFE_STAGE = "biolink:LifeStage"
    PATHOLOGICAL_PROCESS = "biolink:PathologicalProcess"
    DRUG = "biolink:Drug"
    SMALL_MOLECULE = "biolink:SmallMolecule"
    ORGANISM_TAXON = "biolink:OrganismTaxon"
    INFORMATION_CONTENT_ENTITY = "biolink:InformationContentEntity"
    NUCLEIC_ACID_ENTITY = "biolink:NucleicAcidEntity"
    EVIDENCE_TYPE = "biolink:EvidenceType"
    RNAPRODUCT = "biolink:RNAProduct"
    TRANSCRIPT = "biolink:Transcript"
    FUNGUS = "biolink:Fungus"
    PLANT = "biolink:Plant"
    PROCESSED_MATERIAL = "biolink:ProcessedMaterial"
    ACTIVITY = "biolink:Activity"
    AGENT = "biolink:Agent"
    CONFIDENCE_LEVEL = "biolink:ConfidenceLevel"
    DATASET = "biolink:Dataset"
    ENVIRONMENTAL_FEATURE = "biolink:EnvironmentalFeature"
    GENETIC_INHERITANCE = "biolink:GeneticInheritance"
    HAPLOTYPE = "biolink:Haplotype"
    INVERTEBRATE = "biolink:Invertebrate"
    MAMMAL = "biolink:Mammal"
    POPULATION_OF_INDIVIDUAL_ORGANISMS = "biolink:PopulationOfIndividualOrganisms"
    PROTEIN_FAMILY = "biolink:ProteinFamily"
    PUBLICATION = "biolink:Publication"
    ACCESSIBLE_DNA_REGION = "biolink:AccessibleDnaRegion"
    BACTERIUM = "biolink:Bacterium"
    BIOLOGICAL_SEX = "biolink:BiologicalSex"
    CELL_LINE = "biolink:CellLine"
    CHEMICAL_EXPOSURE = "biolink:ChemicalExposure"
    CHEMICAL_MIXTURE = "biolink:ChemicalMixture"
    DATASET_DISTRIBUTION = "biolink:DatasetDistribution"
    DIAGNOSTIC_AID = "biolink:DiagnosticAid"
    DRUG_EXPOSURE = "biolink:DrugExposure"
    ENVIRONMENTAL_PROCESS = "biolink:EnvironmentalProcess"
    EVENT = "biolink:Event"
    EXON = "biolink:Exon"
    GENOME = "biolink:Genome"
    GENOTYPIC_SEX = "biolink:GenotypicSex"
    HUMAN = "biolink:Human"
    INDIVIDUAL_ORGANISM = "biolink:IndividualOrganism"
    MATERIAL_SAMPLE = "biolink:MaterialSample"
    MICRO_RNA = "biolink:MicroRNA"
    ORGANISMAL_ENTITY = "biolink:OrganismalEntity"
    PATENT = "biolink:Patent"
    PHENOTYPIC_SEX = "biolink:PhenotypicSex"
    POLYPEPTIDE = "biolink:Polypeptide"
    PROTEIN_DOMAIN = "biolink:ProteinDomain"
    REAGENT_TARGETED_GENE = "biolink:ReagentTargetedGene"
    REGULATORY_REGION = "biolink:RegulatoryRegion"
    SI_RNA = "biolink:SiRNA"
    SNV = "biolink:Snv"
    STUDY = "biolink:Study"
    STUDY_VARIABLE = "biolink:StudyVariable"
    TRANSCRIPTION_FACTOR_BINDING_SITE = "biolink:TranscriptionFactorBindingSite"
    TREATMENT = "biolink:Treatment"
    WEB_PAGE = "biolink:WebPage"
    ZYGOSITY = "biolink:Zygosity"


class AssociationCategory(Enum):
    """Association categories"""

    PAIRWISE_GENE_TO_GENE_INTERACTION = "biolink:PairwiseGeneToGeneInteraction"
    GENE_TO_EXPRESSION_SITE_ASSOCIATION = "biolink:GeneToExpressionSiteAssociation"
    VARIANT_TO_GENE_ASSOCIATION = "biolink:VariantToGeneAssociation"
    MACROMOLECULAR_MACHINE_TO_BIOLOGICAL_PROCESS_ASSOCIATION = (
        "biolink:MacromolecularMachineToBiologicalProcessAssociation"
    )
    GENE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:GeneToPhenotypicFeatureAssociation"
    MACROMOLECULAR_MACHINE_TO_MOLECULAR_ACTIVITY_ASSOCIATION = (
        "biolink:MacromolecularMachineToMolecularActivityAssociation"
    )
    MACROMOLECULAR_MACHINE_TO_CELLULAR_COMPONENT_ASSOCIATION = (
        "biolink:MacromolecularMachineToCellularComponentAssociation"
    )
    ASSOCIATION = "biolink:Association"
    GENE_TO_GENE_HOMOLOGY_ASSOCIATION = "biolink:GeneToGeneHomologyAssociation"
    GENOTYPE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:GenotypeToPhenotypicFeatureAssociation"
    DISEASE_TO_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:DiseaseToPhenotypicFeatureAssociation"
    GENE_TO_PATHWAY_ASSOCIATION = "biolink:GeneToPathwayAssociation"
    DISEASE_OR_PHENOTYPIC_FEATURE_TO_LOCATION_ASSOCIATION = "biolink:DiseaseOrPhenotypicFeatureToLocationAssociation"
    CHEMICAL_TO_PATHWAY_ASSOCIATION = "biolink:ChemicalToPathwayAssociation"
    GENOTYPE_TO_DISEASE_ASSOCIATION = "biolink:GenotypeToDiseaseAssociation"
    CORRELATED_GENE_TO_DISEASE_ASSOCIATION = "biolink:CorrelatedGeneToDiseaseAssociation"
    DISEASE_OR_PHENOTYPIC_FEATURE_TO_GENETIC_INHERITANCE_ASSOCIATION = (
        "biolink:DiseaseOrPhenotypicFeatureToGeneticInheritanceAssociation"
    )
    VARIANT_TO_DISEASE_ASSOCIATION = "biolink:VariantToDiseaseAssociation"
    CAUSAL_GENE_TO_DISEASE_ASSOCIATION = "biolink:CausalGeneToDiseaseAssociation"
    CHEMICAL_TO_DISEASE_OR_PHENOTYPIC_FEATURE_ASSOCIATION = "biolink:ChemicalToDiseaseOrPhenotypicFeatureAssociation"


class AssociationPredicate(Enum):
    """Association predicates"""

    INTERACTS_WITH = "biolink:interacts_with"
    EXPRESSED_IN = "biolink:expressed_in"
    HAS_PHENOTYPE = "biolink:has_phenotype"
    IS_SEQUENCE_VARIANT_OF = "biolink:is_sequence_variant_of"
    ENABLES = "biolink:enables"
    ACTIVELY_INVOLVED_IN = "biolink:actively_involved_in"
    ORTHOLOGOUS_TO = "biolink:orthologous_to"
    LOCATED_IN = "biolink:located_in"
    SUBCLASS_OF = "biolink:subclass_of"
    RELATED_TO = "biolink:related_to"
    PARTICIPATES_IN = "biolink:participates_in"
    ACTS_UPSTREAM_OF_OR_WITHIN = "biolink:acts_upstream_of_or_within"
    ACTIVE_IN = "biolink:active_in"
    PART_OF = "biolink:part_of"
    MODEL_OF = "biolink:model_of"
    CAUSES = "biolink:causes"
    ACTS_UPSTREAM_OF = "biolink:acts_upstream_of"
    HAS_MODE_OF_INHERITANCE = "biolink:has_mode_of_inheritance"
    CONTRIBUTES_TO = "biolink:contributes_to"
    GENE_ASSOCIATED_WITH_CONDITION = "biolink:gene_associated_with_condition"
    TREATS_OR_APPLIED_OR_STUDIED_TO_TREAT = "biolink:treats_or_applied_or_studied_to_treat"
    COLOCALIZES_WITH = "biolink:colocalizes_with"
    ACTS_UPSTREAM_OF_POSITIVE_EFFECT = "biolink:acts_upstream_of_positive_effect"
    ACTS_UPSTREAM_OF_OR_WITHIN_POSITIVE_EFFECT = "biolink:acts_upstream_of_or_within_positive_effect"
    ACTS_UPSTREAM_OF_NEGATIVE_EFFECT = "biolink:acts_upstream_of_negative_effect"
    ACTS_UPSTREAM_OF_OR_WITHIN_NEGATIVE_EFFECT = "biolink:acts_upstream_of_or_within_negative_effect"


class MappingPredicate(Enum):
    """Mapping predicates"""

    EXACT_MATCH = "skos:exactMatch"
    BROAD_MATCH = "skos:broadMatch"



from urllib.parse import urlencode
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from monarch_py.utils.utils import escape
from pydantic import BaseModel, Field


class core(Enum):
    ENTITY = "entity"
    ASSOCIATION = "association"
    SSSOM = "sssom"


class HistoPhenoKeys(Enum):
    skeletal_system = "UPHENO:0002964"  # "HP:0000924"
    nervous_system = "UPHENO:0004523"  # "HP:0000707"
    head_neck = "UPHENO:0002764"  # "HP:0000152"
    integument = "UPHENO:0002635"  # "HP:0001574"
    eye = "UPHENO:0003020"  # "HP:0000478"
    cardiovascular_system = "UPHENO:0080362"  # "HP:0001626"
    metabolism_homeostasis = "HP:0001939"  # ??? No uPheno parent
    genitourinary_system = "UPHENO:0002642"  # "HP:0000119"
    digestive_system = "UPHENO:0002833"  # "HP:0025031"
    neoplasm = "HP:0002664"  # ??? No uPheno parent
    blood = "UPHENO:0004459"  # "HP:0001871"
    immune_system = "UPHENO:0002948"  # "HP:0002715"
    endocrine = "UPHENO:0003116"  # "HP:0000818"
    musculature = "UPHENO:0002816"  # "HP:0003011"
    respiratory = "UPHENO:0004536"  # "HP:0002086"
    ear = "HP:0000598"  # UPHENO:0002903
    connective_tissue = "UPHENO:0002712"  # "HP:0003549"
    prenatal_or_birth = "UPHENO:0075949"  # "HP:0001197"
    growth = "UPHENO:0049874"  # "HP:0001507"
    breast = "UPHENO:0003013"  # "HP:0000769"


class SolrQuery(BaseModel):
    q: str = "*:*"
    rows: int = 20
    start: int = 0
    facet: bool = True
    facet_min_count: int = 1
    facet_fields: Optional[List[str]] = Field(default_factory=list)
    facet_queries: Optional[List[str]] = Field(default_factory=list)
    filter_queries: Optional[List[str]] = Field(default_factory=list)
    facet_mincount: int = 1
    query_fields: Optional[str] = None
    def_type: str = "edismax"
    q_op: str = "AND"  # See SOLR-8812, need this plus mm=100% to allow boolean operators in queries
    mm: str = "100%"  # All tokens in the query must be found in the doc
    boost: Optional[str] = None
    sort: Optional[str] = None

    def add_field_filter_query(self, field: str, value: Union[list, str, None]):
        if not value or len(value) == 0:
            return self
        if isinstance(value, list):
            fq = " OR ".join([f"{field}:{escape(val)}" for val in value])
        else:
            fq = f"{field}:{escape(value)}"
        self.filter_queries.append(fq)
        return self

    def add_filter_query(self, filter_query: Union[list, str, None]):
        if not filter_query:
            return self
        if isinstance(filter_query, list):
            filter_query = " OR ".join(escape(filter_query))
        self.filter_queries.append(filter_query)
        return self

    def query_string(self):
        return urlencode(
            {self._solrize(k): self._solrize(v) for k, v in self.model_dump().items() if v is not None},
            doseq=True,
        )

    def _solrize(self, value):
        """
        Rename fields and values as necessary to go from the python API to solr query syntax
        """
        if value == "facet_fields":
            return "facet.field"
        elif value == "facet_queries":
            return "facet.query"
        elif value == "filter_queries":
            return "fq"
        elif value == "facet_mincount":
            return "facet.mincount"
        elif value == "query_fields":
            return "qf"
        elif value == "def_type":
            return "defType"
        elif value == "q_op":
            return "q.op"
        elif value is True:
            return "true"
        elif value is False:
            return "false"
        else:
            return value


class SolrQueryResponseHeader(BaseModel):
    QTime: int
    params: Any


class SolrQueryResponse(BaseModel):
    num_found: int = Field(alias="numFound")
    start: int
    docs: List[Any]


class SolrFacetCounts(BaseModel):
    facet_fields: Optional[Dict]
    facet_queries: Optional[Dict]


class SolrQueryResult(BaseModel):
    responseHeader: SolrQueryResponseHeader
    response: SolrQueryResponse
    facet_counts: Optional[SolrFacetCounts]



from typing import List, Union

from fastapi import APIRouter, Depends, Query, Response

from monarch_py.api.additional_models import OutputFormat, PaginationParams
from monarch_py.api.config import solr
from monarch_py.datamodels.model import AssociationResults, CompactAssociationResults, MultiEntityAssociationResults
from monarch_py.datamodels.category_enums import AssociationCategory, AssociationPredicate, EntityCategory
from monarch_py.utils.format_utils import to_tsv

router = APIRouter(
    tags=["association"],
    responses={404: {"description": "Not Found"}},
)


@router.get("")
@router.get("/all", include_in_schema=False)  # We can remove this once the chatgpt plugin & oak aren't using it
async def _get_associations(
    category: List[AssociationCategory] = Query(default_factory=list),
    subject: Union[List[str], None] = Query(default=None),
    subject_category: List[EntityCategory] = Query(default_factory=list),
    subject_namespace: Union[List[str], None] = Query(default=None),
    subject_taxon: Union[List[str], None] = Query(default=None),
    predicate: List[AssociationPredicate] = Query(default_factory=list),
    object: Union[List[str], None] = Query(default=None),
    object_category: List[EntityCategory] = Query(default_factory=list),
    object_namespace: Union[List[str], None] = Query(default=None),
    object_taxon: Union[List[str], None] = Query(default=None),
    entity: Union[List[str], None] = Query(default=None),
    direct: bool = Query(default=False),
    facet_fields: List[str] = Query(default_factory=list),
    facet_queries: List[str] = Query(default_factory=list),
    filter_queries: List[str] = Query(default_factory=list),
    compact: bool = Query(default=False),
    pagination: PaginationParams = Depends(),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
) -> Union[AssociationResults, CompactAssociationResults, str]:
    """Retrieves all associations for a given entity, or between two entities."""
    response = solr().get_associations(
        category=category,
        subject=subject,
        predicate=predicate,
        object=object,
        entity=entity,
        subject_category=subject_category,
        subject_namespace=subject_namespace,
        subject_taxon=subject_taxon,
        object_taxon=object_taxon,
        object_category=object_category,
        object_namespace=object_namespace,
        direct=direct,
        compact=compact,
        facet_fields=facet_fields,
        facet_queries=facet_queries,
        filter_queries=filter_queries,
        offset=pagination.offset,
        limit=pagination.limit,
    )
    if format == OutputFormat.json:
        return response
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")


@router.get("/multi", include_in_schema=False)
async def _get_multi_entity_associations(
    entity: Union[List[str], None] = Query(default=None),
    counterpart_category: Union[List[str], None] = Query(default=None),
    pagination: PaginationParams = Depends(),
    format: OutputFormat = Query(
        default=OutputFormat.json,
        title="Output format for the response",
        examples=["json", "tsv"],
    ),
) -> List[MultiEntityAssociationResults]:
    """Retrieves all associations between each entity and each counterpart category."""
    response = solr().get_multi_entity_associations(
        entity=entity,
        counterpart_category=counterpart_category,
        offset=pagination.offset,
        limit_per_group=pagination.limit,
    )
    if format == OutputFormat.json:
        return response
    elif format == OutputFormat.tsv:
        tsv = ""
        for row in to_tsv(response, print_output=False):
            tsv += row
        return Response(content=tsv, media_type="text/tab-separated-values")



from typing import List, Any
import requests

from pydantic import BaseModel

from monarch_py.api.additional_models import SemsimMetric, SemsimMultiCompareRequest, SemsimDirectionality
from monarch_py.datamodels.model import TermSetPairwiseSimilarity, SemsimSearchResult, Entity


class SemsimianService(BaseModel):
    """A class that makes HTTP requests to the semsimian_server."""

    semsim_server_port: int
    semsim_server_host: str
    entity_implementation: Any  # TODO: should be EntityInterface

    def convert_tsps_data(self, data):
        """Convert to a format that can be coerced into a TermSetPairwiseSimilarity model

        TODO: currently, the response returned from semsimian_server doesn't
        100% match the TermSetPairwiseSimilarity model, so we perform some
        transformations below. once it does, we can remove all the code below
        and just return TermSetPairwiseSimilarity(**data)
        """
        # remove these similarity maps and fold them into the _best_matches dicts
        object_best_matches_similarity_map = self._convert_nans(data.pop("object_best_matches_similarity_map"))
        subject_best_matches_similarity_map = self._convert_nans(data.pop("subject_best_matches_similarity_map"))
        converted_data = {
            **data,
            **{
                # flatten the nested termset dicts
                "subject_termset": {k: v for d in data["subject_termset"] for k, v in d.items()},
                "object_termset": {k: v for d in data["object_termset"] for k, v in d.items()},
                "subject_best_matches": {
                    k: {**v, "similarity": subject_best_matches_similarity_map[k]}
                    for k, v in data["subject_best_matches"].items()
                },
                "object_best_matches": {
                    k: {**v, "similarity": object_best_matches_similarity_map[k]}
                    for k, v in data["object_best_matches"].items()
                },
            },
        }
        return converted_data

    def compare(
        self, subjects: List[str], objects: List[str], metric: SemsimMetric = SemsimMetric.ANCESTOR_INFORMATION_CONTENT
    ) -> TermSetPairwiseSimilarity:
        host = f"http://{self.semsim_server_host}:{self.semsim_server_port}"
        path = f"compare/{','.join(subjects)}/{','.join(objects)}/{metric}"
        url = f"{host}/{path}"

        print(f"Fetching {url}...")
        response = requests.get(url=url)
        data = response.json()
        results = self.convert_tsps_data(data)
        return TermSetPairwiseSimilarity(**results)

    def multi_compare(self, request: SemsimMultiCompareRequest) -> List[SemsimSearchResult]:
        comparison_results = [
            self.compare(request.subjects, object_set.phenotypes, request.metric) for object_set in request.object_sets
        ]
        results = [
            SemsimSearchResult(
                subject=Entity(id=object_set.id, name=object_set.label),
                score=comparison_result.average_score,
                similarity=comparison_result,
            )
            for object_set, comparison_result in zip(request.object_sets, comparison_results)
        ]
        return results

    def search(
        self,
        termset: List[str],
        prefix: str,
        metric: SemsimMetric = SemsimMetric.ANCESTOR_INFORMATION_CONTENT,
        directionality: SemsimDirectionality = SemsimDirectionality.BIDIRECTIONAL,
        limit: int = 10,
    ) -> List[SemsimSearchResult]:
        host = f"http://{self.semsim_server_host}:{self.semsim_server_port}"
        path = f"search/{','.join(termset)}/{prefix}:/{metric}?limit={limit}&directionality={directionality.value} "
        url = f"{host}/{path}"

        print(f"Fetching {url}...")
        response = requests.get(url=url)
        data = response.json()
        results = [
            SemsimSearchResult(
                subject=self.entity_implementation.get_entity(i[2], extra=False),
                score=i[0],
                similarity=self.convert_tsps_data(i[1]),
            )
            for i in data
        ]

        return results

    @staticmethod
    def _convert_nans(input_dict, to_value=None):
        """
        Given an input dict of the form {<term>: {<field>: <value>, ...}}
        converts any <value> of 'NaN' to None.
        """
        for k, v in input_dict.items():
            for ik, iv in v.items():
                if iv == "NaN":
                    input_dict[k][ik] = None

        return input_dict



from fastapi import APIRouter, Query, Response
from typing import List

from monarch_py.api.additional_models import TextAnnotationRequest
from monarch_py.api.config import spacyner
from monarch_py.datamodels.model import TextAnnotationResult

router = APIRouter(tags=["text_annotation"], responses={404: {"description": "Not Found"}})


@router.get("/annotate")
def _annotate(text: str = Query(default="", title="The text content to annotate")) -> str:
    return Response(content=spacyner().annotate_text(text), media_type="text/html")


@router.post("/annotate")
def _post_annotate(request: TextAnnotationRequest) -> str:
    return _annotate(request.content)


@router.get("/annotate/entities")
def _entities(text: str = Query(default="", title="")) -> List[TextAnnotationResult]:
    return spacyner().get_annotated_entities(text)


@router.post("/annotate/entities")
def _post_entities(request: TextAnnotationRequest) -> List[TextAnnotationResult]:
    return _entities(request.content)
